use crate::{DBError, DBResult};

use std::{
    collections::HashMap,
    sync::{Arc, Mutex},
};

pub type DBConnection = std::sync::Arc<Connection>;

pub(crate) struct CachedStatement {
    stmt: sqlite::Statement<'static>,
    sql: String,
}

pub(crate) trait PreparedKey {
    fn into_u64(self) -> u64;
}

impl PreparedKey for u64 {
    fn into_u64(self) -> u64 {
        self
    }
}

impl PreparedKey for std::any::TypeId {
    fn into_u64(self) -> u64 {
        use std::hash::Hash;
        use std::hash::Hasher;
        let mut hasher = std::collections::hash_map::DefaultHasher::new();
        self.hash(&mut hasher);
        hasher.finish()
    }
}

pub struct Connection {
    // we leak the ConnectionThreadSafe and make sure that the only references to it are stored in
    // statement_cache, so as long as we drop the statement_cache first there are no correctness
    // issues, and we get to have static-lifetime statements for caching purposes.
    conn: &'static sqlite::ConnectionThreadSafe,
    statement_cache: Mutex<HashMap<u64, CachedStatement>>,
}

/*
#[cfg(test)]
mod sendsync_check {
    struct CheckSend<T: Send>(std::marker::PhantomData<T>);
    struct CheckSync<T: Send>(std::marker::PhantomData<T>);

    #[test]
    fn check_send() {
        let cs = CheckSend::<super::DBConnection>(Default::default());
    }
}
*/

impl Connection {
    pub fn open<U: AsRef<str>>(uri: U) -> Result<DBConnection, DBError> {
        match sqlite::Connection::open_thread_safe_with_flags(
            uri.as_ref(),
            sqlite::OpenFlags::new()
                .with_create()
                .with_full_mutex()
                .with_read_write(),
        ) {
            Ok(conn) => {
                let conn = Box::leak(Box::new(conn));
                Ok(Arc::new(Self {
                    conn,
                    statement_cache: Default::default(),
                }))
            }
            Err(e) => Err(DBError::Sqlite(e)),
        }
    }

    pub(crate) fn execute_raw(&self, sql: &str) -> Result<(), DBError> {
        Ok(self.conn.execute(sql)?)
    }

    pub(crate) fn with_prepared<R>(
        &self,
        hash_key: impl PreparedKey,
        build_query: impl Fn() -> String,
        run_query: impl Fn(&mut sqlite::Statement<'static>) -> DBResult<R>,
    ) -> DBResult<R> {
        match self.statement_cache.lock()?.entry(hash_key.into_u64()) {
            std::collections::hash_map::Entry::Vacant(e) => {
                let sql = build_query();

                let q: sqlite::Statement<'static> = self
                    .conn
                    .prepare(sql.as_str())
                    .map_err(|e| DBError::from(e).sqlite_to_query(sql.as_str()))?;

                log::trace!("prepared new SQL query: {sql}");

                let inserted = e.insert(CachedStatement {
                    sql: sql.clone(),
                    stmt: q,
                });
                run_query(&mut inserted.stmt).map_err(|e| e.sqlite_to_query(sql.as_str()))
            }
            std::collections::hash_map::Entry::Occupied(mut e) => run_query(&mut e.get_mut().stmt)
                .map_err(|err| err.sqlite_to_query(&e.get_mut().sql)),
        }
    }
}

impl Drop for Connection {
    fn drop(&mut self) {
        // first we drop the statement cache, if we can get it
        // if this fails then we're entering UD, but we probably panicked anyway
        self.statement_cache.lock().map(|mut sc| sc.clear()).ok();
        // then we drop the connection
        unsafe {
            drop(Box::from_raw(
                self.conn as *const _ as *mut sqlite::ConnectionThreadSafe,
            ));
        }
    }
}

impl std::fmt::Debug for Connection {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("microrm::Connection")
    }
}
