//! Component types for query construction.

use crate::{
    prelude::Queryable,
    query::{AssocInterface, QueryPart},
    schema::{
        datum::{Datum, DatumList, DatumListRef, DatumVisitor},
        entity::{Entity, EntityPart, EntityPartList, EntityPartVisitor},
        DatumDiscriminator, IDMap, IDWrap,
    },
};

use super::Query;

/// Concrete implementation of Queryable for an IDMap
pub(crate) struct MapQueryable<'a, E: Entity> {
    map: &'a IDMap<E>,
}

impl<'a, E: Entity> MapQueryable<'a, E> {
    pub fn new(map: &'a IDMap<E>) -> Self {
        Self { map }
    }
}

impl<'a, E: Entity> Queryable for MapQueryable<'a, E> {
    type EntityOutput = E;
    type OutputContainer = Vec<IDWrap<E>>;
    type StaticVersion = MapQueryable<'static, E>;

    fn build(&self) -> Query {
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT".into())
            .attach(QueryPart::Columns, "*".into())
            .attach(QueryPart::From, format!("`{}`", E::entity_name()))
    }
    fn bind(&self, _stmt: &mut sqlite::Statement, _index: &mut usize) {}

    fn conn(&self) -> &crate::db::DBConnection {
        self.map.conn()
    }
}

/// Concrete implementation of Queryable for an IDMap
pub(crate) struct AssocQueryable<'a, AI: AssocInterface> {
    assoc: &'a AI,
}

impl<'a, AI: AssocInterface> AssocQueryable<'a, AI> {
    pub fn new(assoc: &'a AI) -> Self {
        Self { assoc }
    }
}

impl<'a, AI: AssocInterface> Queryable for AssocQueryable<'a, AI> {
    type EntityOutput = AI::RemoteEntity;
    type OutputContainer = Vec<IDWrap<AI::RemoteEntity>>;
    type StaticVersion = AssocQueryable<'static, AI>;

    fn build(&self) -> Query {
        let anames = super::AssocNames::collect(self.assoc).unwrap();
        let assoc_name = anames.assoc_name();
        Query::new()
            .attach(QueryPart::Root, "SELECT DISTINCT".into())
            .attach(QueryPart::Columns, format!("`{}`.*", anames.remote_name))
            .attach(QueryPart::From, format!("`{}`", assoc_name))
            .attach(
                QueryPart::Join,
                format!(
                    "`{}` ON `{}`.`id` = `{}`.`{}`",
                    anames.remote_name, anames.remote_name, assoc_name, anames.remote_field
                ),
            )
            .attach(
                QueryPart::Where,
                format!("`{}`.`{}` = ?", assoc_name, anames.local_field),
            )
    }
    fn bind(&self, stmt: &mut sqlite::Statement, index: &mut usize) {
        let adata = self
            .assoc
            .get_data()
            .expect("binding query for assoc with no data");

        stmt.bind((*index, adata.local_id))
            .expect("couldn't bind assoc id");
        *index += 1;
    }

    fn conn(&self) -> &crate::db::DBConnection {
        &self.assoc.get_data().unwrap().conn
    }
}

/// Filter on a Datum
pub(crate) struct WithComponent<'a, WEP: EntityPart, Parent: Queryable> {
    datum: &'a WEP::Datum,
    parent: Parent,
    _ghost: std::marker::PhantomData<WEP>,
}

impl<'a, WEP: EntityPart, Parent: Queryable> WithComponent<'a, WEP, Parent> {
    pub fn new(parent: Parent, _part: WEP, datum: &'a WEP::Datum) -> Self {
        Self {
            datum,
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<'a, WEP: EntityPart, Parent: Queryable> Queryable for WithComponent<'a, WEP, Parent> {
    type EntityOutput = WEP::Entity;
    type OutputContainer = Parent::OutputContainer;
    type StaticVersion = WithComponent<'static, WEP, Parent::StaticVersion>;

    fn build(&self) -> Query {
        self.parent.build().attach(
            QueryPart::Where,
            format!(
                "`{}`.`{}` = ?",
                <Self::EntityOutput>::entity_name(),
                WEP::part_name()
            ),
        )
    }
    fn bind(&self, stmt: &mut sqlite::Statement, index: &mut usize) {
        self.parent.bind(stmt, index);
        self.datum.bind_to(stmt, *index);
        *index += 1;
    }

    fn conn(&self) -> &crate::db::DBConnection {
        self.parent.conn()
    }
}

/// Filter on the unique index
pub(crate) struct UniqueComponent<'a, E: Entity, Parent: Queryable> {
    datum: <<E::Uniques as EntityPartList>::DatumList as DatumList>::Ref<'a>,
    parent: Parent,
}

impl<'a, E: Entity, Parent: Queryable> UniqueComponent<'a, E, Parent> {
    pub fn new(
        parent: Parent,
        datum: <<E::Uniques as EntityPartList>::DatumList as DatumList>::Ref<'a>,
    ) -> Self {
        Self { datum, parent }
    }
}

impl<'a, E: Entity, Parent: Queryable> Queryable for UniqueComponent<'a, E, Parent> {
    type EntityOutput = E;
    type OutputContainer = Option<IDWrap<E>>;
    type StaticVersion = UniqueComponent<'static, E, Parent::StaticVersion>;

    fn build(&self) -> Query {
        let mut query = self.parent.build();

        struct PartVisitor<'a>(&'a mut Query);
        impl<'a> EntityPartVisitor for PartVisitor<'a> {
            fn visit<EP: EntityPart>(&mut self) {
                self.0.attach_mut(
                    QueryPart::Where,
                    format!(
                        "`{}`.`{}` = ?",
                        <EP::Entity>::entity_name(),
                        EP::part_name()
                    ),
                );
            }
        }

        <E::Uniques>::accept_part_visitor(&mut PartVisitor(&mut query));

        query
    }

    fn bind(&self, stmt: &mut sqlite::Statement, index: &mut usize) {
        self.parent.bind(stmt, index);

        struct Visitor<'a, 'b>(&'a mut sqlite::Statement<'b>, &'a mut usize);
        impl<'a, 'b> DatumVisitor for Visitor<'a, 'b> {
            fn visit<ED: Datum>(&mut self, datum: &ED) {
                datum.bind_to(self.0, *self.1);
                *self.1 += 1;
            }
        }

        self.datum.accept(&mut Visitor(stmt, index));
    }

    fn conn(&self) -> &crate::db::DBConnection {
        self.parent.conn()
    }
}

pub(crate) struct SingleComponent<Parent: Queryable> {
    parent: Parent,
}

impl<Parent: Queryable> SingleComponent<Parent> {
    pub fn new(parent: Parent) -> Self {
        Self { parent }
    }
}

impl<Parent: Queryable> Queryable for SingleComponent<Parent> {
    type EntityOutput = Parent::EntityOutput;
    type OutputContainer = Option<IDWrap<Self::EntityOutput>>;
    type StaticVersion = SingleComponent<Parent::StaticVersion>;

    fn build(&self) -> Query {
        self.parent
            .build()
            .attach(QueryPart::Trailing, "LIMIT 1".into())
    }

    fn bind(&self, stmt: &mut sqlite::Statement, index: &mut usize) {
        self.parent.bind(stmt, index)
    }

    fn conn(&self) -> &crate::db::DBConnection {
        self.parent.conn()
    }
}

/// Join with another entity via an association
pub(crate) struct JoinComponent<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable>
{
    parent: Parent,
    _ghost: std::marker::PhantomData<(L, R, EP)>,
}

impl<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable>
    JoinComponent<R, L, EP, Parent>
{
    pub fn new(parent: Parent, _part: EP) -> Self {
        Self {
            parent,
            _ghost: Default::default(),
        }
    }
}

impl<R: Entity, L: Entity, EP: EntityPart<Entity = L>, Parent: Queryable> Queryable
    for JoinComponent<R, L, EP, Parent>
{
    type EntityOutput = R;
    type OutputContainer = Vec<IDWrap<R>>;
    type StaticVersion = JoinComponent<R, L, EP, Parent::StaticVersion>;

    fn build(&self) -> Query {
        let remote_name = R::entity_name();
        let local_name = L::entity_name();
        let part_name = EP::part_name();
        let assoc_name = format!("{local_name}_{remote_name}_assoc_{part_name}");

        struct Discriminator(Option<(&'static str, &'static str)>);
        impl DatumDiscriminator for Discriminator {
            fn visit_entity_id<E: Entity>(&mut self) {
                unreachable!()
            }
            fn visit_serialized<T: serde::Serialize + serde::de::DeserializeOwned>(&mut self) {
                unreachable!()
            }
            fn visit_bare_field<T: Datum>(&mut self) {
                unreachable!()
            }

            fn visit_assoc_map<E: Entity>(&mut self) {
                self.0 = Some(("domain", "range"));
            }
            fn visit_assoc_domain<R: crate::schema::Relation>(&mut self) {
                self.0 = Some(("domain", "range"));
            }
            fn visit_assoc_range<R: crate::schema::Relation>(&mut self) {
                self.0 = Some(("range", "domain"));
            }
        }

        let mut d = Discriminator(None);
        <EP::Datum>::accept_discriminator(&mut d);

        let (local_field, remote_field) = d.0.unwrap();

        self.parent
            .build()
            .attach(
                QueryPart::Join,
                format!("`{assoc_name}` ON `{local_name}`.`id` = `{assoc_name}`.`{local_field}`"),
            )
            .attach(
                QueryPart::Join,
                format!(
                    "`{remote_name}` ON `{assoc_name}`.`{remote_field}` = `{remote_name}`.`id`"
                ),
            )
            .replace(QueryPart::Columns, format!("`{remote_name}`.*"))
    }
    fn bind(&self, stmt: &mut sqlite::Statement, index: &mut usize) {
        self.parent.bind(stmt, index);
    }

    fn conn(&self) -> &crate::db::DBConnection {
        self.parent.conn()
    }
}

/*
SQL mapping:

MapQueryable<E>::get()
 -> SELECT * FROM {E::entity_name()}

UniqueComponent<E, MapQueryable<E>>::get()
 -> SELECT * FROM {E::entity_name()} WHERE `col1` = ? AND `col2` = ? AND ...

WithComponent<E, EP, MapQueryable<E>>::get()
 -> SELECT * FROM {E::entity_name()} WHERE `col1` = ?

AssocQueryable<AI>::get()
 -> SELECT
        `{AI::RemoteEntity::entity_name()}`.*
    FROM `{assoc_table_name}`
    LEFT JOIN `{AI::RemoteEntity::entity_name()}`.id = `{assoc_table_name}`.`{remote_field}`
    WHERE `{assoc_table_name}`.`{local_field}` = ?

UniqueComponent<E, AssocQueryable<AI>>::get()
 -> SELECT
        `{AI::RemoteEntity::entity_name()}`.*
    FROM `{assoc_table_name}`
    LEFT JOIN `{AI::RemoteEntity::entity_name()}`.id = `{assoc_table_name}`.`{remote_field}`
    WHERE
        `{assoc_table_name}`.`{local_field}` = ?
        AND `{AI::RemoteEntity::entity_name()}`.`col1` = ?
        AND `{AI::RemoteEntity::entity_name()}`.`col2` = ?
        ...

JoinComponent<R, E, EP, MapQueryable<E>>::get()
 -> SELECT DISTINCT
        `{R::entity_name()}`.*
    FROM
        `{E::entity_name()}`
    LEFT JOIN `{assoc_table_name}` ON `{E::entity_name()}`.`id` = `{assoc_table_name}`.`{local_field}`
    LEFT JOIN `{R::entity_name()}` ON `{assoc_table_name}`.`{remote_field}` = `{R::entity_name()}`.`id`

*/
